/* 
 * Exemple de navigation utilisant gluLookAt et illustrant le
 * mipmaping (touche m).  (exemple fait en cours avec les tudiants)
 * Auteur : Fares BELHADJ (Lire COPYING) Email :
 * amsi@ai.univ-paris8.fr Date : 25/10/2013
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

#define EPSILON 0.015

static SDL_Surface * sdlInit(int w, int h);
static void          sdlInitGL(SDL_Surface * s);
static void          sdlResizeGL(SDL_Surface * s);
static void          sdlLoop(SDL_Surface * s);
static int           sdlManageEvents(SDL_Surface * s);
static void          sdlDraw(SDL_Surface * s);
static void          cube(void);
static void          plan(void);
static void          sdlSphere(int onlyFreeData);
static inline void   sdlFPS(void);

static int pause = 0;
static GLuint texId[3] = {0, 0, 0}, curId = 0;

static GLdouble ma[3] = {0, -9.8, 0};
static GLdouble mv[3] = {0, 0, 0};
static GLdouble mp[3] = {0, 2, 0};


enum {
  KEY_UP = 0,
  KEY_RIGHT,
  KEY_DOWN,
  KEY_LEFT,
};

static int keys[4]  = {0, 0, 0, 0};

static GLfloat cam[4] = {0, 0, 0, 0}; /* x, y, z, thta (orientation vers +z, augmente  droite, diminue  gauche) */

static SDL_Surface * texSurface = NULL;

/* un tout petit main */
int main(int argc, char ** argv) {
  SDL_Surface * s;
  s = sdlInit(640, 480);
  sdlLoop(s);
  sdlSphere(1);
  return 0;
}

/* sdlInit : initialiser SDL pour OpenGL */
static SDL_Surface * sdlInit(int w, int h) {
  SDL_Surface * s;
  const SDL_VideoInfo * vi = NULL;
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Impossible d'initialiser SDL: %s\n", SDL_GetError());
    exit(1);
  }

  if( (vi = SDL_GetVideoInfo()) == 0 ) {
    fprintf(stderr, "Echec de SDL_GetVideoInfo(): %s\n", SDL_GetError());
    exit(1);
  }
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE,     8 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE,   8 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE,    8 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE,  16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  if ( (s = SDL_SetVideoMode(w, h, vi->vfmt->BitsPerPixel,
			     SDL_OPENGL | SDL_RESIZABLE)) == NULL ) {
    fprintf(stderr, "Impossible d'ouvrir le mode video : %s\n", SDL_GetError());
    exit(1);
  }
  atexit(SDL_Quit);
  sdlInitGL(s);
  return s;
}

/*
 * Pour celles et ceux qui n'ont pas accs  la constante 
 * GL_BGR dans glTexImage2D
static void bgr2rgb(unsigned char * data, int size) {
  int i;
  unsigned char c;
  for (i = 0; i < size; i += 3) {
      c = data[i];
      data[i] = data[i + 2];
      data[i + 2] = c;
  }
}
*/

/* sdlInitGL : configurer les options GL */
static void sdlInitGL(SDL_Surface * s) {
  GLfloat pos[] = {0, 300, 100, 1};
  GLfloat amb[] = {0.5, 0.5, 0.5, 1};
  GLfloat dif[] = {1, 1, 0.8, 1};
  glShadeModel(GL_SMOOTH);
  glDepthFunc(GL_LEQUAL);
  glFrontFace(GL_CCW);
  glCullFace(GL_BACK);
  glClearColor(0, 0, 0, 0);
  glClearDepth(1.0);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  
  glLightfv(GL_LIGHT0, GL_POSITION, pos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, dif);
  glLightfv(GL_LIGHT0, GL_AMBIENT, amb);

  if( (texSurface = SDL_LoadBMP("tex01.bmp")) == NULL ) {
    fprintf(stderr, "Impossible d'ouvrir le fichier : %s\n", SDL_GetError());
    exit(1);
  }
  /*bgr2rgb(texSurface->pixels, 3 * texSurface->w * texSurface->h);*/

  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);

  glGenTextures(3, texId);

  glBindTexture(GL_TEXTURE_2D, texId[0]); 
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, texSurface->w, texSurface->h, 0, GL_BGR, GL_UNSIGNED_BYTE, texSurface->pixels);

  glBindTexture(GL_TEXTURE_2D, texId[1]); 
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, texSurface->w, texSurface->h, 0, GL_BGR, GL_UNSIGNED_BYTE, texSurface->pixels);
  
  glBindTexture(GL_TEXTURE_2D, texId[2]); 
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, texSurface->w, texSurface->h, 0, GL_BGR, GL_UNSIGNED_BYTE, texSurface->pixels);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3, texSurface->w, texSurface->h, GL_BGR, GL_UNSIGNED_BYTE, texSurface->pixels);
  
  glEnable(GL_TEXTURE_2D);

  SDL_FreeSurface(texSurface);

  sdlResizeGL(s);

}

/* 
 * CallBack : en cas de resize modifier la matrice de projection en
 * fonction du rapport w/h
 */
static void sdlResizeGL(SDL_Surface * s) {
  glViewport(0, 0, s->w, s->h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, s->w / (float)s->h, 0.1, 100.0);
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );
}

static void sdlSimu(void) {
  double dt;
  Uint32 t;
  static Uint32 t0 = 0;
  t = SDL_GetTicks();
  dt = (t - t0) / 1000.0;
  t0 = t;
  if(keys[KEY_UP]) {
    cam[0] += - dt * 10 * sin(cam[3]);
    cam[2] +=   dt * 10 * cos(cam[3]);
  }
  if(keys[KEY_DOWN]) {
    cam[0] -= - dt * 10 * sin(cam[3]);
    cam[2] -=   dt * 10 * cos(cam[3]);
  }
  if(keys[KEY_LEFT])
    cam[3] -= dt * 1 * M_PI;
  if(keys[KEY_RIGHT])
    cam[3] += dt * 1 * M_PI;

  mv[0] += ma[0] * dt;
  mv[1] += ma[1] * dt;
  mv[2] += ma[2] * dt;

  mp[0] += mv[0] * dt;
  mp[1] += mv[1] * dt;
  mp[2] += mv[2] * dt;

  if(mp[1] <= 0 && mv[1] < 0) {
    mp[1] = 0;
    mv[1] = -(1.0 + EPSILON) * mv[1];
  }
}

/* 
 * loop principale : sdlManageEvents renvoie vrai si la fenetre est
 * inactive => tantque inactive dors une milliseconde.
 * sinon dessiner puis afficher le framerate
 */
static void sdlLoop(SDL_Surface * s) {
  for(;;) {
    while(sdlManageEvents(s)) SDL_Delay(1);
    sdlSimu();
    sdlDraw(s);
    sdlFPS();
  }
}

/* 
 * sdlManageEvents gere les evenements ; aussi renvoie vrai si la
 * fenetre est inactive, sinon faux.
 * Les evenements sont : 
 *
 * escape, q, Q : quitter
 *
 * directions : se dplacer dans l'univers
 *
 * p : pause
 *
 * m : changer de filtrage pour la texture
 *
 * w : activer / desactiver le mode wireframe
 *
 * l : activer / desactiver l'eclairage
 *
 * f : activer / desactiver le dgrad
 *
 * F1 : fullscreen On / Off
 * 
 * redimensionnement de fenetre : appeler sdlResizeGL
 *
 */
static int sdlManageEvents(SDL_Surface * s) {
  GLint v[2];
  static int active = 1;
  SDL_Event event;
  while(SDL_PollEvent(&event)) 
    switch (event.type) {
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case SDLK_UP:
	keys[KEY_UP] = 1;
	break;
      case SDLK_DOWN:
	keys[KEY_DOWN] = 1;
	break;
      case SDLK_LEFT:
	keys[KEY_LEFT] = 1;
	break;
      case SDLK_RIGHT:
	keys[KEY_RIGHT] = 1;
	break;
      case ' ':
      case 'p':
	pause = !pause;
	break;
      case 'm':
	curId = (curId + 1) % 3;
	break;
      case 'w':
	glGetIntegerv(GL_POLYGON_MODE, v);
	if(v[0] == GL_FILL)
	  glPolygonMode(GL_FRONT, GL_LINE);
	else
	  glPolygonMode(GL_FRONT, GL_FILL);
	break;
      case 'f':
	glGetIntegerv(GL_SHADE_MODEL, v);
	if(v[0] == GL_FLAT)
	  glShadeModel(GL_SMOOTH);
	else
	  glShadeModel(GL_FLAT);
	break;
      case 'l':
	if(glIsEnabled(GL_LIGHTING)) {
	  glDisable(GL_LIGHTING);
	  glDisable(GL_LIGHT0);
	} else {
	  glEnable(GL_LIGHTING);
	  glEnable(GL_LIGHT0);
	}
	break;
      case SDLK_ESCAPE:
      case 'q':
	exit(0);
      case SDLK_F1:
	SDL_WM_ToggleFullScreen(s);
	break;
      default:
	fprintf(stderr, "La touche %s a ete pressee\n",
		SDL_GetKeyName(event.key.keysym.sym));
	break;
      }
      break;
    case SDL_KEYUP:
      switch(event.key.keysym.sym) {
      case SDLK_UP:
	keys[KEY_UP] = 0;
	break;
      case SDLK_DOWN:
	keys[KEY_DOWN] = 0;
	break;
      case SDLK_LEFT:
	keys[KEY_LEFT] = 0;
	break;
      case SDLK_RIGHT:
	keys[KEY_RIGHT] = 0;
	break;
      default: break;
      }
      break;
    case SDL_ACTIVEEVENT:
      if(event.active.state & SDL_APPACTIVE)
	active = event.active.gain;
      break;
    case SDL_VIDEORESIZE:
      if( (s = SDL_SetVideoMode(event.resize.w, event.resize.h, 
				s->format->BitsPerPixel, s->flags)) == NULL ) {
	fprintf(stderr, "Echec de SDL_SetVideoMode apres le resize : %s\n", 
		SDL_GetError());
	exit(1);
      }
      sdlResizeGL(s);
      break;
    case SDL_QUIT:
      exit(0);
    }
  return !active;
}

static void sdlDraw(SDL_Surface * s) {
  Uint32 t;
  float dt;
  int xm, ym, i;
  static float a = 0.0;
  static Uint32 t0 = 0;
  static GLfloat data[] = { -0.25, 0.25, 0.25, 0.25, 0.25, -0.25, -0.25, -0.25 };
  GLfloat mat[16], nord[3];
  t = SDL_GetTicks();
  dt = (t - t0) / 1000.0;
  t0 = t;

  SDL_PumpEvents();
  SDL_GetMouseState(&xm, &ym);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();

  glBindTexture(GL_TEXTURE_2D, texId[curId]);
  gluLookAt(cam[0], cam[1], cam[2], cam[0] - sin(cam[3]), cam[1]-(ym - (s->h >> 1)) / (float)(s->h), cam[2] + cos(cam[3]), 0, 1, 0);
  glGetFloatv(GL_MODELVIEW_MATRIX, mat);
  for(i = 8; i < 11; i++)
    nord[i - 8] = mat[i];
  nord[2] = -nord[2];

  glPushMatrix();
  glTranslatef(3, 0, 6);
  glRotatef(a, 0, 1, 0);
  cube();
  glPopMatrix();
  
  glPushMatrix();
  glTranslatef(0, 0, 6);
  glTranslatef(mp[0], mp[1], mp[2]);

  glPushAttrib(GL_ENABLE_BIT);
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);

  sdlSphere(0);

  glPopAttrib();

  glPopMatrix();
  
  plan();

  glPushAttrib(GL_ENABLE_BIT);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(-1, 1, -1, 1, 0.0, 100);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glColor3f(0, 1, 0);

  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(2, GL_FLOAT, 0, data);

  glPushMatrix();
  glTranslatef(-0.5, -0.5, 0);
  glRotatef(a, 0, 0, 1);
  glDrawArrays(GL_LINE_LOOP, 0, 4);
  glPopMatrix();

  glPushMatrix();
  glTranslatef(0.5, -0.5, 0);
  glRotatef(-a, 0, 0, 1);
  glDrawArrays(GL_LINE_LOOP, 0, 4);
  glPopMatrix();

  glDisableClientState(GL_VERTEX_ARRAY);

  glPushMatrix();
  glTranslatef(0.5, 0.5, 0);
  glBegin(GL_LINES);
  glColor3f(1, 0, 0);
  glVertex2f(0, 0);
  glVertex2f(0.3 * nord[0], 0.3 * nord[2]);
  glColor3f(1, 1, 1);
  glVertex2f(0, 0);
  glVertex2f(-0.3 * nord[0], -0.3 * nord[2]);
  glEnd();
  glPopMatrix();

  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);

  glPopAttrib();

  if(!pause)
    a += 360 * dt;
  SDL_GL_SwapBuffers();
}


/*\!brief dessine une sphre unitaire colorie ou ombre (vecteurs
 * normaux) ; cette fonction utilise des QUADS.
 *
 * \param onlyFreeData demande  la fonction de librer sa donne sans
 * rien mettre  GL.
 */
static void sdlSphere(int onlyFreeData) {
  static GLfloat * vertices = NULL;
  int i, e;
  GLfloat phi0, phi1, theta0, theta1;  
  static int longiMax = 30, latiMax = 30;
  if(onlyFreeData) {
    if(vertices) {
      free(vertices);
      vertices = NULL;
    }
    return;
  }
  if(!vertices) {
    vertices = malloc(2 * latiMax * 2 * longiMax * 3 * sizeof *vertices);
    assert(vertices);
  }
  for(e = 0 ; e < latiMax; e++) {
    theta0 = -(M_PI / 2) + (e * M_PI) / (GLfloat)(latiMax);
    theta1 = -(M_PI / 2) + ((e + 1) * M_PI) / (GLfloat)(latiMax);
    for(i = 0; i < longiMax; i++) {
      phi0 = (2.0 * M_PI * i) / (GLfloat)longiMax;
      phi1 = (2.0 * M_PI * (i + 1)) / (GLfloat)longiMax;
      vertices[12 * (longiMax * e + i) + 0] = cos(theta0) * cos(phi0);
      vertices[12 * (longiMax * e + i) + 1] = sin(theta0);
      vertices[12 * (longiMax * e + i) + 2] = cos(theta0) * sin(phi0);
      vertices[12 * (longiMax * e + i) + 3] = cos(theta0) * cos(phi1);
      vertices[12 * (longiMax * e + i) + 4] = sin(theta0);
      vertices[12 * (longiMax * e + i) + 5] = cos(theta0) * sin(phi1);
      vertices[12 * (longiMax * e + i) + 6] = cos(theta1) * cos(phi1);
      vertices[12 * (longiMax * e + i) + 7] = sin(theta1);
      vertices[12 * (longiMax * e + i) + 8] = cos(theta1) * sin(phi1);
      vertices[12 * (longiMax * e + i) + 9] = cos(theta1) * cos(phi0);
      vertices[12 * (longiMax * e + i) + 10] = sin(theta1);
      vertices[12 * (longiMax * e + i) + 11] = cos(theta1) * sin(phi0);
    }
  }
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);
  glVertexPointer(3, GL_FLOAT, 0, vertices);
  glNormalPointer(GL_FLOAT, 0, vertices);
  glColorPointer(3, GL_FLOAT, 0, vertices);
  glDrawArrays(GL_QUADS, 0, 2 * 2 * longiMax * latiMax);
  glDisableClientState(GL_COLOR_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
}

static void cube(void) {
  GLfloat data[] = { /* 3 pour normales, 2 pour coord de texture, 3 pour sommet */
    /* Arriere */ 
    0,  0, -1, 1,  0, -1,  1, -1, 0,  0, -1, 0,  0, 1,  1, -1, 0,  0, -1, 0, 1, 1, -1, -1, 0,  0, -1, 1,  1, -1, -1, -1,
    /* Gauche */
    -1,  0,  0, 1,  1, -1, -1,  1, -1,  0, 0, 1,  0, -1,  1,  1, -1,  0, 0, 0,  0, -1,  1, -1, -1,  0, 0, 0,  1, -1, -1, -1,
    /* Droite */
    1,  0,  0, 0,  1,  1, -1,  1, 1,  0, 0, 1,  1,  1, -1, -1, 1,  0, 0, 1,  0,  1,  1, -1, 1,  0, 0, 0,  0,  1,  1,  1,
    /* Avant */
    0,  0,  1, 0,  0, -1,  1,  1, 0,  0, 1, 0,  1, -1, -1,  1, 0,  0, 1, 1,  1,  1, -1,  1, 0,  0, 1, 1,  0,  1,  1,  1,
    /* Haut */
    0,  1,  0, 0,  1, -1,  1,  1, 0,  1, 0, 1,  1,  1,  1,  1, 0,  1, 0, 1,  0,  1,  1, -1, 0,  1, 0, 0,  0, -1,  1, -1,
    /* Bas */
    0, -1,  0, 1,  1, -1, -1,  1, 0,  -1, 0, 1,  0, -1, -1, -1, 0,  -1, 0, 0,  0,  1, -1, -1, 0,  -1, 0, 0,  1,  1, -1,  1 };
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glNormalPointer(GL_FLOAT, 8 * sizeof *data, data);
  glTexCoordPointer(2, GL_FLOAT, 8 * sizeof *data, &data[3]);
  glVertexPointer(3, GL_FLOAT, 8 * sizeof *data, &data[5]);
  glDrawArrays(GL_QUADS, 0, 24);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
}

static void plan(void) {
  GLfloat data[] = { /* 3 pour normales, 2 pour coord de texture, 3 pour sommet */
    0, 1, 0, 20,  20, -20, -1,  20, 
    0, 1, 0,  0,  20,  20, -1,  20, 
    0, 1, 0,  0,   0,  20, -1, -20, 
    0, 1, 0, 20,   0, -20, -1, -20
  };
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glNormalPointer(GL_FLOAT, 8 * sizeof *data, data);
  glTexCoordPointer(2, GL_FLOAT, 8 * sizeof *data, &data[3]);
  glVertexPointer(3, GL_FLOAT, 8 * sizeof *data, &data[5]);
  glDrawArrays(GL_QUADS, 0, 4);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
}

/*
 * sdlFPS : compter le nombre de frames affiches. imprimer le
 * framerate toutes les 5 secondes.
 */
static inline void sdlFPS(void) {
  Uint32 t;
  static Uint32 t0 = 0, f = 0;
  f++;
  t = SDL_GetTicks();
  if(t - t0 > 5000) {
    fprintf(stderr, "sdlFPS:%6d\n",
	    (int)(1000.0 * f / (t - t0)));
    t0 = t;
    f  = 0;
  }
}

